/************************************************************
*    parseBitmap.c
*
*    Functions:
*    textureBitmap
*    LoadDIBitmap
*************************************************************/
/************************************************************
*    Code created and written by Chris Hegarty
*    Feel free to use any portion of the code,
*    but I would appreciate credit if you use it.
*************************************************************/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include "bitmap.h"

/************************************************************
*    textureBitmap
*    Usage: textureBitmap(char *, TEXTUREINFO *);
*    Purpose: creates a texture from the bitmap file
*    Returns: Texture number (GLuint)
*************************************************************/

GLuint textureBitmap(char *acFile,TEXTUREINFO *sTextureInfo){

	GLuint iPicTexture;
	void *pPictureArray;
	BITMAPINFO *sInfo;
	
	  /* Parse bitmap and return pointer to data array */
	pPictureArray = LoadDIBitmap(acFile,&sInfo,sTextureInfo);
	if(pPictureArray == NULL)
	    return 0;

      /* Generate 2D texture */
	glGenTextures(1,&iPicTexture);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glBindTexture(GL_TEXTURE_2D,iPicTexture);
	glTexImage2D(GL_TEXTURE_2D,0,3,sTextureInfo->texture_width,
			sTextureInfo->texture_height,0,GL_RGB,GL_UNSIGNED_BYTE,pPictureArray);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	free(pPictureArray);
	free(sInfo);
	return iPicTexture;
}

/************************************************************
*    LoadDIBitmap
*    Usage: LoadDIBitmap(char *, BITMAPINFO *,TEXTUREINFO *);
*    Purpose: parses data from bitmap file
*    Returns: pointer to data array
*************************************************************/

void *LoadDIBitmap(char *acFilename, BITMAPINFO **sInfo, TEXTUREINFO *sTextureInfo){
	FILE *fd;
    char *pBits;
	long lBitsSize, lInfoSize,lTextureSize;
	int i,j;
	char acFileType[3];
	char cTemp;
	char *pTempArray;
	BITMAPFILEHEADER sHeader;
	int iWidth = 1,iHeight = 1;

      /* Open bitmap file */
	if((fd = fopen(acFilename,"rb")) == NULL){
	    MessageBox(NULL,"Could not open bitmap file","Error",MB_OK);
		return NULL;
	}
      /* read bitmap header */
	if(fread(&sHeader, sizeof(BITMAPFILEHEADER),1,fd) < 1){
		MessageBox(NULL,"Could not read bitmap file header","Error",MB_OK);
		fclose(fd);
		return NULL;
	}

	memcpy(acFileType,&sHeader.bfType,2);
	acFileType[2] = '\0';
	  /* Check if bitmap file */
	if(strcmp(acFileType,"BM")){
	   MessageBox(NULL,"File not bitmap","Error",MB_OK);
	   fclose(fd);
	   return NULL;
	}

      /* Read Bitmap info header */
	lInfoSize = sHeader.bfOffBits - sizeof(BITMAPFILEHEADER);
	if((*sInfo = (BITMAPINFO *)malloc(lInfoSize)) == NULL){
		MessageBox(NULL,"Could not allocate memory for bitmap info header","Error",MB_OK);
		fclose(fd);
		return NULL;
	}
	if(fread(*sInfo,1,lInfoSize,fd) < (unsigned long)lInfoSize){
		MessageBox(NULL,"Could not read bitmap info header","Error",MB_OK);
		free(*sInfo);
		fclose(fd);
		return NULL;
	}

      /* Since OpenGL only allows texture data as power of two */
      /* we must determine the nearest larger power of two for */
      /* the bitmap size */
	while(iWidth < (*sInfo)->bmiHeader.biWidth)
		iWidth *= 2;
	while(iHeight < (*sInfo)->bmiHeader.biHeight)
		iHeight *= 2;

	sTextureInfo->image_width = (*sInfo)->bmiHeader.biWidth;
	sTextureInfo->image_height = (*sInfo)->bmiHeader.biHeight;
	sTextureInfo->texture_width = iWidth;
	sTextureInfo->texture_height = iHeight;

	if((lBitsSize = (*sInfo)->bmiHeader.biSizeImage) == 0){
		lBitsSize = ((*sInfo)->bmiHeader.biWidth *
				   (*sInfo)->bmiHeader.biBitCount + 7) / 8 *
				   abs((*sInfo)->bmiHeader.biHeight);
	}
	lTextureSize = iWidth*iHeight*3;

      /* allocate memory for data */
	if((pBits = (char *)malloc(lTextureSize)) == NULL){
		MessageBox(NULL,"Could not allocate memory for bitmap data","Error",MB_OK);
		free(*sInfo);
		fclose(fd);
		return NULL;
	}
      /* allocate a temporary array of equal size */
	if((pTempArray = (char *)malloc(lTextureSize)) == NULL){
		MessageBox(NULL,"Could not allocate memory for bitmap data","Error",MB_OK);
		free(*sInfo);
		free(pBits);
		fclose(fd);
		return NULL;
	}
      /* Read data from file */
	if(fread(pTempArray,1,lBitsSize,fd) < (unsigned long)lBitsSize){
		MessageBox(NULL,"Could not read bitmap data","Error",MB_OK);
		free(*sInfo);
		free(pBits);
		free(pTempArray);
		fclose(fd);
		return NULL;
	}
	

      /* Reverse bytes from bgr to rgb */
	for(i=0;i<lBitsSize;i+=3){
		cTemp = pTempArray[i];
	    pTempArray[i] = pTempArray[i+2];
		pTempArray[i+2] = cTemp;
	}


	  /* Restructure data array to pad for irregular sizes */
	for(i=0;i<(*sInfo)->bmiHeader.biHeight;i++){
		for(j=0;j<(*sInfo)->bmiHeader.biWidth*3;j++){
			pBits[i * iWidth * 3 + j] = 
			               pTempArray[i * (*sInfo)->bmiHeader.biWidth * 3 + j];
		}
	}

	fclose(fd);
	free(pTempArray);
	return(pBits);
}
