/************************************************************
*    particleFountain.cpp
*
*    Functions:
*        particleFountain
*        Particle class
*************************************************************/
/************************************************************
*    Code created and written by Chris Hegarty
*    Feel free to use any portion of the code,
*    but I would appreciate credit if you use it.
*************************************************************/

#include <windows.h>
#include <stdlib.h>
#include <GL/gl.h>
#include "particleFountain.h"

/************************************************************
*    Particle class
*    Purpose: Contains the information and functions for
*             displaying a single fountain particle
*************************************************************/

/************************************************************
*    Particle constructor
*    Usage: Particle particle;
*    Purpose: Creates the particle object
*************************************************************/

Particle::Particle(GravityObject  *object){

	position = object->GetPosition();
	position.y += object->GetRadius();
	velocity = object->GetVelocity();

	velocity.x = velocity.x * velocity.w +
	             (float)rand() / (float) RAND_MAX - .5;
	velocity.y = velocity.y * velocity.w +
	             4.0 * ((MAX_RAND_RADIUS + .1) - object->GetRadius());
	velocity.z = velocity.z * velocity.w +
	             (float)rand() / (float) RAND_MAX - .5;
    velocity = velocity.unit();

    object->numParticles++;
}

/************************************************************
*    Particle destructor
*    Purpose: destroys the particle object
*************************************************************/

Particle::~Particle(void){}

/************************************************************
*    Particle function
*    Usage: particle.Reset();
*    Purpose: Resets the particle's position and velocity to the 
*             owner object.  Allows for particle recycling
*    Returns: Nothing
*************************************************************/

void Particle::Reset(GravityObject  *object){

    position = object->GetPosition();
	position.y += object->GetRadius();
	velocity = object->GetVelocity();

	velocity.x = velocity.x * velocity.w +
	             (float)rand() / (float) RAND_MAX - .5;
	velocity.y = velocity.y * velocity.w +
	             4.0 * ((MAX_RAND_RADIUS + .1) - object->GetRadius());
	velocity.z = velocity.z * velocity.w +
	             (float)rand() / (float) RAND_MAX - .5;
    velocity = velocity.unit();
}

/************************************************************
*    Particle function
*    Usage: particle.pushParticle();
*    Purpose: Places particle pointer on the top of the particle
*             stack.  Determines when a particle expires
*    Returns: Nothing
*************************************************************/

void Particle::pushParticle(GravityObject  *object){
	int i;
	Particle *temp_particle_ptr;

	temp_particle_ptr = this;

	for(i=(object->numParticles)-1;i>0;i--){
		object->particleList[i] = object->particleList[i-1];
	}

	object->particleList[0] = temp_particle_ptr;
}

/************************************************************
*    Particle function
*    Usage: particle.calculateParticleVelocity(GLdouble time);
*    Purpose: Calculates the particles velocity based on
*             acceleration down
*    Returns: Nothing
*************************************************************/

void Particle::calculateParticleVelocity(GLdouble deltaTime){
	Vertex temp;
	temp = velocity;
	temp.y -= GRAVITATION_CONSTANT * deltaTime;
	velocity = temp;
}

/************************************************************
*    Particle function
*    Usage: particle.calculateParticlePostition();
*    Purpose: Calculates the particle's position
*    Returns: Nothing
*************************************************************/

void Particle::calculateParticlePosition(GLdouble deltaTime){
	position.x  += velocity.x * deltaTime;
	position.y  += velocity.y * deltaTime;
	position.z  += velocity.z * deltaTime;
}

/************************************************************
*    particleFountain
*    Usage: particleFountain(GravityObjet *object, GLdouble deltaTime);
*    Purpose: Performs particle operations on all particles in the
*             object's particle stack
*    Returns: success or failure
*************************************************************/

int particleFountain(GravityObject *object,GLdouble deltaTime){
	GLuint i;
	Particle *particle;
	GLfloat position[3];

    if(object->numParticles < MAX_PARTICLES)
        particle = new Particle(object);
	else{
		particle = (Particle *)object->particleList[MAX_PARTICLES-1];
		particle->Reset(object);
    }

	particle->pushParticle(object);
	
	glPushMatrix();
	glPushAttrib(GL_LIGHTING_BIT);
	glMaterialfv(GL_FRONT,GL_EMISSION,object->particleColor);
	glBegin(GL_POINTS);

	for(i=0;i<object->numParticles;i++){
		particle = (Particle *)object->particleList[i];
		particle->calculateParticleVelocity(deltaTime);
        particle->calculateParticlePosition(deltaTime);
		position[0] = particle->position.x;
		position[1] = particle->position.y;
		position[2] = particle->position.z;
		glVertex3fv(position);
	}

	glEnd();
	glPopAttrib();
	glPopMatrix();
	return 0;
}