/************************************************************
*    physLib.c
*
*    Functions:
*        GravityObject class
*        Vertex class
*        Vector class
*        getColor
*        calculateGravityForce
*        calculateAcceleration
*        calculateVelocity
*        calculatePosition
*        calculateCollision
*        calculateMomentum
*        calculateEnergy
*************************************************************/
/************************************************************
*    Code created and written by Chris Hegarty
*    Feel free to use any portion of the code,
*    but I would appreciate credit if you use it.
*************************************************************/
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <math.h>
#include <stdlib.h>
#include "physLib.h"
#include "bitmap.h"

#define NUM_TEXTURES  12
#ifndef M_PI
#define M_PI 3.141592654
#endif

/************************************************************
*    GravityObject class
*    Purpose: Contains the information and functions for
*             controlling the gravity objects
*************************************************************/

/************************************************************
*    GravityObject constructor
*    Usage: GravityObject object;
*    Purpose: Creates and initializes a gravity affected object
*************************************************************/

GravityObject::GravityObject(void){
    float white[4] = {.8,.8,.8,1.0};
    TEXTUREINFO texture_info;
      /* Texture names for object spheres */
    char filename[NUM_TEXTURES][20] = {
                            "bitmap/rock1.bmp",
                            "bitmap/rock3.bmp",
                            "bitmap/rock4.bmp",
                            "bitmap/rock5.bmp",
                            "bitmap/rock6.bmp",
                            "bitmap/water1.bmp",
                            "bitmap/water2.bmp",
                            "bitmap/water3.bmp",
                            "bitmap/metal1.bmp",
                            "bitmap/metal2.bmp",
                            "bitmap/cloud1.bmp",
                            "bitmap/cloud2.bmp",};
      /* Set up quadric for sphere */
    GLUquadricObj *sphere;
    sphere = gluNewQuadric();
    gluQuadricDrawStyle(sphere,GLU_FILL);
    gluQuadricNormals(sphere,GLU_SMOOTH);
    gluQuadricOrientation(sphere,GLU_OUTSIDE);
    gluQuadricTexture(sphere,GL_TRUE);

      /* Set random position for objects */
    position.x = (MAX_RAND_X - MIN_RAND_X) *
                         ((double)rand() / (double)RAND_MAX) + MIN_RAND_X;
    position.y = (MAX_RAND_Y - MIN_RAND_Y) *
                         ((double)rand() / (double)RAND_MAX) + MIN_RAND_Y;
    position.z = (MAX_RAND_Z - MIN_RAND_Z) *
                         ((double)rand() / (double)RAND_MAX) + MIN_RAND_Z;

      /* Set random object radius */
    radius = (MAX_RAND_RADIUS - MIN_RAND_RADIUS) *
             ((double)rand() / (double)RAND_MAX) + 
             MIN_RAND_RADIUS;

      /* Calculate mass based on constant density */
      /* for a sphere, (4/3 PI r^3) */
    mass = DENSITY * (4.0/3.0) * M_PI * radius *
                                radius * radius;
      /* Calculate Inertia based on solid sphere (2/5 M r^2) */
    inertia = (2.0/5.0) * mass * radius * radius;

      /* Get random color for particle fountain */
    getColor(particleColor);

      /* Create a display list for the gravity object */
      /* generate texture */
    texture_map = textureBitmap(filename[(int)((float)rand()/
                 (float)RAND_MAX *(float)NUM_TEXTURES)],&texture_info);

    displayList = glGenLists(1);
    glNewList(displayList,GL_COMPILE);
          /* BIND TEXTURE */
        if(texture_map != 0){
            glBindTexture(GL_TEXTURE_2D,texture_map);
            glEnable(GL_TEXTURE_2D);
        }
          /* Set color */
        glMaterialfv(GL_FRONT,GL_DIFFUSE,white);
          /* Create quadric sphere */
        glRotatef(90,0,1,0);
        glRotatef(90,1,0,0);
        gluSphere(sphere,radius,10,10);
        glDisable(GL_TEXTURE_2D);

    glEndList();

      /* clear number of fountain particles */
    numParticles = 0;
}

/************************************************************
*    GravityObject destructor
*    Purpose: Destroys the gravity object
*************************************************************/

GravityObject::~GravityObject(void){}

/************************************************************
*    GravityObject function
*    Usage: object.SetAcceleration(Vector accel);
*    Purpose: Sets the acceleration vector for the object
*    Returns: nothing
*************************************************************/

void GravityObject::SetAcceleration(Vector accel_in){
    acceleration = accel_in;
}

/************************************************************
*    GravityObject accessor
*    Usage: object.GetAcceleration();
*    Purpose: Get the acceleration vector for the object
*    Returns: Vector
*************************************************************/

Vector GravityObject::GetAcceleration(void){
    return acceleration;
}

/************************************************************
*    GravityObject function
*    Usage: object.SetVelocity(Vector vel);
*    Purpose: Sets the velocity vector for the object
*    Returns: nothing
*************************************************************/

void GravityObject::SetVelocity(Vector vel_in){
    velocity = vel_in;
}

/************************************************************
*    GravityObject accessor
*    Usage: object.GetVelocity();
*    Purpose: Get the velocity vector for the object
*    Returns: Vector
*************************************************************/

Vector GravityObject::GetVelocity(void){
    return velocity;
}

/************************************************************
*    GravityObject function
*    Usage: object.SetPosition(Vertex pos);
*    Purpose: Sets the position vertex for the object
*    Returns: nothing
*************************************************************/

void GravityObject::SetPosition(Vertex pos_in){
    position = pos_in;
}

/************************************************************
*    GravityObject accessor
*    Usage: object.GetPosition();
*    Purpose: Get the position vertex for the object
*    Returns: Vertex class
*************************************************************/

Vertex GravityObject::GetPosition(void){
    return position;
}

/************************************************************
*    GravityObject function
*    Usage: object.SetRadius(GLdouble radius);
*    Purpose: Sets the radius for the object
*    Returns: nothing
*************************************************************/

void GravityObject::SetRadius(GLdouble rad_in){
    radius = rad_in;
}

/************************************************************
*    GravityObject accessor
*    Usage: object.GetRadius();
*    Purpose: Get the radius for the object
*    Returns:GLdouble
*************************************************************/

GLdouble GravityObject::GetRadius(void){
    return radius;
}

/************************************************************
*    GravityObject function
*    Usage: object.SetMass(GLdouble mass);
*    Purpose: Sets the mass for the object
*    Returns: nothing
*************************************************************/

void GravityObject::SetMass(GLdouble mass_in){
    mass = mass_in;
}

/************************************************************
*    GravityObject accessor
*    Usage: object.GetMass();
*    Purpose: Get the mass for the object
*    Returns:GLdouble
*************************************************************/

GLdouble GravityObject::GetMass(void){
    return mass;
}

/************************************************************
*    Vertex class
*    Purpose: Contains the information and functions for
*             controlling the 3 element vertex 
*************************************************************/

/************************************************************
*    Vertex constructor
*    Usage: Vertex vertex;
*    Purpose: Creates and initializes a vertex object with all 0's
*************************************************************/

Vertex::Vertex(void){
    x = 0; y = 0; z = 0;
}

/************************************************************
*    Vertex constructor
*    Usage: Vertex vertex(GLdouble x,GLdouble y,GLdouble z);
*    Purpose: Creates and initializes a vertex object with 
*             specified values
*************************************************************/

Vertex::Vertex(GLdouble x1, GLdouble y1, GLdouble z1){
    x= x1; y = y1; z =z1;
}

/************************************************************
*    Vertex destructor
*    Purpose: Destroys the Vertex object
*************************************************************/

Vertex::~Vertex(){}

/************************************************************
*    Vertex operator
*    Usage: vertex + vertex;
*    Purpose: overloads addition to add all three elements
*************************************************************/

Vertex Vertex::operator+(Vertex vertex2){
    Vertex sum;

    sum.x = x + vertex2.x;
    sum.y = y + vertex2.y;
    sum.z = z + vertex2.z;
    
    return sum;
}

/************************************************************
*    Vertex operator
*    Usage: vertex - vertex;
*    Purpose: overloads subtraction to subtract all three elements
*************************************************************/

Vertex Vertex::operator-(Vertex vertex2){
    Vertex difference;

    difference.x = x - vertex2.x;
    difference.y = y - vertex2.y;
    difference.z = z - vertex2.z;
    return difference;
}

/************************************************************
*    Vertex operator
*    Usage: vertex * vertex;
*    Purpose: overloads multiplication to return dot product
*************************************************************/

GLdouble Vertex::operator*(Vertex vertex2){
    GLdouble output;
    output = x * vertex2.x + y * vertex2.y + z * vertex2.z;
    return output;
}

/************************************************************
*    Vertex operator
*    Usage: vertex * GLdouble;
*    Purpose: overloads multiply for scalar multiplication
*************************************************************/

Vertex Vertex::operator*(GLdouble scalar){
    Vertex output;
    output.x = x * scalar;
    output.y = y * scalar;
    output.z = z * scalar;
    return output;
}

/************************************************************
*    Vertex operator
*    Usage: vertex = Vector;
*    Purpose: overloads equal to convert from vector
*************************************************************/

Vertex &Vertex::operator=( Vector vector){
    x = vector.x * vector.w;
    y = vector.y * vector.w;
    z = vector.z * vector.w;

    return *this;
}

/************************************************************
*    Vertex function
*    Usage: vertex.magnitude();
*    Purpose: return the magnitude of the vertex
*    Returns: GLdouble
*************************************************************/

GLdouble Vertex::magnitude(void){
    GLdouble mag;
    mag = sqrt(x*x + y*y + z*z);
    return mag;
}

/************************************************************
*    Vertex function
*    Usage: vertex.unit();
*    Purpose: return the unit "vector" of the vertex
*    Returns: Vertex
*************************************************************/

Vertex Vertex::unit(void){
    GLdouble mag;
    Vertex unit;

    mag = magnitude();
    if(mag == 0.0){
        unit.x = 0.0;
        unit.y = 0.0;
        unit.z = 0.0;
    }
    else{
        unit.x = x / mag;
        unit.y = y / mag;
        unit.z = z / mag;
    }

    return unit;
}

/************************************************************
*    Vector class
*    Purpose: Contains the information and functions for
*             controlling the 4 element vector 
*************************************************************/

/************************************************************
*    Vector constructor
*    Usage: Vector vector;
*    Purpose: Creates and initializes a vector object with all 0's
*************************************************************/

Vector::Vector(void){
    x = 0; y = 0; z = 0; w = 0;
}

/************************************************************
*    Vector constructor
*    Usage: Vector vector(x,y,z,w);
*    Purpose: Creates and initializes a vector object with
*             specified values
*************************************************************/

Vector::Vector(GLdouble x1, GLdouble y1, GLdouble z1, 
                                             GLdouble w1){
    x= x1; y = y1; z =z1; w = w1;
}

/************************************************************
*    Vector destructor
*    Purpose: Destroys vector object
*************************************************************/

Vector::~Vector(){}

/************************************************************
*    Vector operator
*    Usage: Vector + Vector;
*    Purpose: overloads addition for vector addition
*************************************************************/

Vector Vector::operator+(Vector vector2){
    Vector sum;

    if(w == 0.0)
        return vector2;

    sum.x = x * w + vector2.x * vector2.w;
    sum.y = y * w + vector2.y * vector2.w;
    sum.z = z * w + vector2.z * vector2.w;
    sum = sum.unit();

    return sum;
}

/************************************************************
*    Vector operator
*    Usage: Vector - Vector;
*    Purpose: overloads subtraction for vector subtraction
*************************************************************/

Vector Vector::operator-(Vector vector2){
    Vector difference;

    difference.x = x * w - vector2.x * vector2.w;
    difference.y = y * w - vector2.y * vector2.w;
    difference.z = z * w - vector2.z * vector2.w;
    difference = difference.unit();

    return difference;
}

/************************************************************
*    Vector operator
*    Usage: Vector * Vector;
*    Purpose: overloads multiplication for vector dot product
*    Returns: GLdouble
*************************************************************/

GLdouble Vector::operator*(Vector vector2){
    GLdouble output;
    output = x * w * vector2.x * vector2.w +
             y * w * vector2.y * vector2.w +
             z * w * vector2.z * vector2.w;

    return output;
}

/************************************************************
*    Vector operator
*    Usage: Vector * GLdouble;
*    Purpose: overloads multiplication for vector times scalar
*    Returns: Vector
*************************************************************/

Vector Vector::operator*(GLdouble scalar){
    Vector product;
    product.x = x;
    product.y = y;
    product.z = z;
    product.w = w * scalar;
    return product;
}

/************************************************************
*    Vector operator
*    Usage: Vector = vertex;
*    Purpose: overloads equal for converting vertex to vector
*    Returns: Vector
*************************************************************/

Vector &Vector::operator=(Vertex vertex){
    x = vertex.x;
    y = vertex.y;
    z = vertex.z;
    *this = unit();

    return *this;
}

/************************************************************
*    Vector function
*    Usage: vector.magnitude();
*    Purpose: returns the magnitude of the vector
*    Returns: GLdouble
*************************************************************/

GLdouble Vector::magnitude(void){
    GLdouble mag;
    mag = sqrt(x*x + y*y + z*z);
    return mag;
}

/************************************************************
*    Vector function
*    Usage: vector.unit();
*    Purpose: returns the unit vector
*    Returns: Vector
*************************************************************/

Vector Vector::unit(void){
    GLdouble mag;
    Vector unit;

    mag = magnitude();
    if(mag == 0.0){
        unit.x = 0.0;
        unit.y = 0.0;
        unit.z = 0.0;
    }
    else{
        unit.x = x / mag;
        unit.y = y / mag;
        unit.z = z / mag;
    }
    unit.w = mag;

    return unit;
}


/************************************************************
*    getColor
*    Usage: getColor(float *);
*    Purpose: returns a random color
*    Returns: Nothing
*************************************************************/

void getColor(float *color){
    color[3]=1.0;

      /* red */
    color[0] = (float)((float)rand()/(float)RAND_MAX);
      /* green */
    color[1] = (float)((float)rand()/(float)RAND_MAX);
      /* blue */
    color[2] = (float)((float)rand()/(float)RAND_MAX);

    return;
}

/************************************************************
*    calculateGravityForce
*    Usage: calculateGravityForce(GravityObject *, GravityObject *);
*    Purpose: calculates the gravitational attraction between the
*             two gravity objects passed to it
*    Returns: A 4 element force vector
*************************************************************/

Vector calculateGravityForce(GravityObject *object1, GravityObject *object2){
    Vector forceVector;

      /* Calculate vector direction by subtracting the position
         of the two objects, returned as a unit vector */
    forceVector = (object1->GetPosition() - object2->GetPosition());

      /* Calculate the magnitude of the gravity vector from Newton's law */
      /* of gravitation */
      /* Use a constant radius^2 because the distance is */
      /* small and the variation in gravity is minimal */
      /* This also allows us to avoid erroneous force vectors */
      /* when one object is inside another */
      /* f = -(G * M1 * M2 / r^2) */
    forceVector.w = -(GRAVITATION_CONSTANT * object1->GetMass() * 
                                             object2->GetMass());/* / 1.0; */

    return forceVector;
}

/************************************************************
*    calculateAcceleration
*    Usage: calculateAcceleration(Vector, GLdouble, GLdouble);
*    Purpose: calculates the accleration of the object based on
*             gravitational force vector;
*    Returns: A 4 element acceleration vector
*************************************************************/

Vector calculateAcceleration(Vector forceVector,GLdouble mass, GLdouble deltaTime){
    Vector acceleration;

      /* get the gravity force vector direction */
    acceleration = forceVector;
      /* calculate the acceleration as the force divided by the mass */
      /* times the elapsed time */
      /* a = (g / M1) * dT */
    acceleration.w = (forceVector.w / mass) * deltaTime;

    return acceleration;
}

/************************************************************
*    calculateVelocity
*    Usage: calculateVelocity(Vector, Vector, GLdouble);
*    Purpose: calculates the velocity of the object based on
*             the acceleration vector;
*    Returns: A 4 element velocity vector
*************************************************************/

Vector calculateVelocity(Vector acceleration,Vector initVelocity,GLdouble deltaTime){
    Vector velocity;


      /* velocity = the previous velocity plus acceleration times */
      /* the elapsed time */
      /* v = a * dT */
    velocity = initVelocity + (acceleration * deltaTime);

    return velocity;
}

/************************************************************
*    calculatePosition
*    Usage: calculatePosition(Vector, Vertex, GLdouble);
*    Purpose: calculates the position of the object based on
*             the velocity vector;
*    Returns: A 3 element position
*************************************************************/

Vertex calculatePosition(Vector velocity,Vertex initPosition,GLdouble deltaTime){
    Vertex position;

      /* position = the previous position plus velocity times */
      /* the elapsed time */
      /* p = v * dT */
    position.x = initPosition.x + velocity.x * velocity.w * deltaTime;
    position.y = initPosition.y + velocity.y * velocity.w * deltaTime;
    position.z = initPosition.z + velocity.z * velocity.w * deltaTime;

    return position;
}

/************************************************************
*    calculateCollision
*    Usage: calculateCollision(GravityObject *, GravityObject *);
*    Purpose: calculates the new velocity vector of objects that collide;
*    Returns: A 4 element velocity vector
*************************************************************/

Vector calculateCollision(GravityObject *object1, GravityObject *object2){
    Vector relativeVelocity;
    Vertex temp_vel;
    Vector newVelocity;
    Vertex positionVector;
    GLdouble j,j_num,j_den;

      /* Post collision velocities are calculated as follows */
      /* V2 = V1 + (j/Ma)N */
      /* j = -(1 + e) Vrel_a_b dot N   */
      /*     -----------------------   */
      /*     N dot N (1/Ma + 1/Mb)     */

      /* Since we make sure N is a unit vector, we can get rid */
      /* of the N dot N in the denominator */

      /* This equation was derived for Chris Hecker's article, */
      /* Physics 3: Collision Response */


      /* Find the vector between the two colliding object centers */
    positionVector = (object1->GetPosition() - object2->GetPosition()).unit();
      /* Find the relative velocity between the two objects */
    relativeVelocity = object1->GetVelocity() - object2->GetVelocity();

      /* calculate the collision coeficient j */
      /* multiply the relative velocity by -( 1 + elasticity) */
    relativeVelocity.w *= -(double)(1.0+ELASTICITY);
      /* the numerator = a dot product between the relative velocity and */
      /* the position vector */
    temp_vel = relativeVelocity;
    j_num = temp_vel * positionVector;
      /* the denominator = ( 1/ Ma + 1 / Mb) */
    j_den = ((1.0 / object1->GetMass()) + (1.0 / object2->GetMass()));
      /* calculate j */
    j = j_num / j_den;
      /* the velocity modifier is a vector = j/Ma times the position vector */
      /* this vector is returned and added to the current velocity */
    newVelocity = positionVector * (j/object1->GetMass());
      /* An unknown constant is added based on the mass of the other object */
      /* The colliding objects somehow "bleed" energy at a very specific rate */
      /* during collision */
    newVelocity.w += UNKNOWN_VELOCITY_CONSTANT * object2->GetMass();

    return newVelocity;

}

/************************************************************
*    calculateMomentum
*    Usage: calculateMomentum(GravityObject *);
*    Purpose: calculates the momentum of the object
*    Returns: A 4 element momentum vector
*************************************************************/

Vector calculateMomentum(GravityObject *object){
    Vector momentum;

      /* momentum = mass times velocity */
    momentum = object->GetVelocity();
    momentum.w *= object->GetMass();
    object->momentum = momentum;

    return object->momentum;
}

/************************************************************
*    createBackground
*    Usage: createBackground(char *);
*    Purpose: creates a display list of a raster graphics bitmap
*    Returns: Display List
*************************************************************/

GLuint createBackground(char *filename){
    TEXTUREINFO sTextureInfo;
    void *pPictureArray;
    BITMAPINFO *sInfo;
    GLfloat xoffset,yoffset;
    GLfloat xsize,ysize;
    GLuint list;


    pPictureArray = LoadDIBitmap(filename,&sInfo,&sTextureInfo);

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    list = glGenLists(1);
    glNewList(list,GL_COMPILE);

    glDrawPixels(sTextureInfo.texture_width,sTextureInfo.texture_height,
                                  GL_RGB,GL_UNSIGNED_BYTE,pPictureArray);

    glEndList();
    free(pPictureArray);
    return list;
}
