/************************************************************
*    phys_eng.c
*
*    Functions:
*        windowInit
*        physInit
*        mainDisplay
*        mainReshape
*        mainMouse
*        mainMotion
*        mainKey
*        clear_hook
*        frameMonitor
*        Gravity List class
*************************************************************/
/************************************************************
*    Code created and written by Chris Hegarty
*    Feel free to use any portion of the code,
*    but I would appreciate credit if you use it.
*************************************************************/

#include <windows.h>
#include <GL/gl.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include "phys_eng.h"
#include "physlib.h"
#include "particleFountain.h"
#include "win_funcs.h"

static GLdouble windim;
static GLuint light_list;
static Vertex CenterOfMass;
const double WINDOW_ZOOM = 9.0;
GLuint background;

static GravityList MasterList(5);
static GravityObject TempObject[MAX_GRAVITY_OBJ];
GLubyte CollisionMatrix[MAX_GRAVITY_OBJ][MAX_GRAVITY_OBJ];

/************************************************************
*    windowInit
*    Usage: windowInit();
*    Purpose: Initializes openGL related elements
*    Returns: Nothing
*************************************************************/

void windowInit(void){
      // Initialize color and position of lights
    GLfloat grey[4] = {.1,.1,.1,1};
    GLfloat bright_white[4] = {.9,.9,.9,1.0};
    GLfloat white[4] = {.9,.9,.9,1};
    GLfloat light0Position[4] = {-1,1,1,1};

      // seed random number generator
    srand(time(NULL));

      // initialize GL state
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_NORMALIZE);
    glCullFace(GL_BACK);
    glEnable(GL_CULL_FACE);
    glShadeModel(GL_SMOOTH);

    glClearColor(0,0,0,0.0);

    glEnable(GL_LIGHTING);

      // create light list, with two white lights
    light_list = glGenLists(1);
    glNewList(light_list,GL_COMPILE);

    glLightfv(GL_LIGHT0,GL_DIFFUSE,white);
    glLightfv(GL_LIGHT0,GL_AMBIENT,grey);
    glLightfv(GL_LIGHT0,GL_SPECULAR,bright_white);
    glLightfv(GL_LIGHT0,GL_POSITION,light0Position);

    glEnable(GL_LIGHT0);

    glEndList();
      /* set point size for particle fountain */
    glPointSize(1.0);

    background = createBackground("bitmap/background.bmp");
}

/************************************************************
*    physEngInit
*    Usage: physEngInit();
*    Purpose: Initializes engine related elements
*    Returns: Nothing
*************************************************************/
void physEngInit(void){
    GLuint i,j;
    double mass_sum=0;

      /* Create gravity objects and find center of mass */
    for(i=0;i<MasterList.GetMaxObjects();i++){
        GravityObject *return_object = new GravityObject();
        MasterList.AddObject(return_object);
        CenterOfMass.x += return_object->GetPosition().x*return_object->GetMass();
        CenterOfMass.y += return_object->GetPosition().y*return_object->GetMass();
        CenterOfMass.z += return_object->GetPosition().z*return_object->GetMass();
        mass_sum += return_object->GetMass();
    }
    CenterOfMass.x /= mass_sum;
    CenterOfMass.y /= mass_sum;
    CenterOfMass.z /= mass_sum;

      /* Initialize collision matrix which indicates what objects are colliding */
    for(i=0;i<MasterList.GetNumObjects();i++){
        for(j=0;j<MasterList.GetNumObjects();j++){
            CollisionMatrix[i][j] = FALSE;
        }
    }
}

/************************************************************
*    mainDisplay
*    Usage: mainDisplay();
*    Purpose: performs the main graphics display update by
*             calling light list, landscape and tree lists,
*             and placing all active objects
*    Returns: Nothing
*************************************************************/
void mainDisplay(void){
    static int last_time = 0;
    static GLdouble averageTime = 0;
    static long int averageCounter = 0;
    int current_time;
    SYSTEMTIME time;
    GLuint i,j;
    GLdouble deltaTime;
    GravityObject *current_obj;

      /* Using elapsed time to update position allows for a relatively */
      /* constant velocity of the balls regardless of processor speed */
    GetSystemTime(&time);
    if(last_time == 0){
        last_time = time.wMilliseconds + (((time.wDay * 24 + time.wHour) * 60 +
                            time.wMinute) * 60 + time.wSecond) * 1000;
        GetSystemTime(&time);
    }
    current_time = time.wMilliseconds + (((time.wDay * 24 + time.wHour) * 60 +
                            time.wMinute) * 60 + time.wSecond) * 1000;
    deltaTime = (double)(current_time - last_time) / 1000.0;
    averageTime = (averageTime * averageCounter + deltaTime) / (averageCounter + 1);
    averageCounter++;
    deltaTime = averageTime;
    deltaTime *= 1.5;
    last_time = current_time;

      /* Calculate objects' new velocity and position */
    mainIdle(deltaTime);

      /* Clear Color buffer.  Depth Buffer isn't cleared in order to speed up */
      /* frame rate on slower machines (like mine).  We can get away with it */
      /* because we keep everything at the same depth */
    glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();

    glPushMatrix();
          /* set up screen*/
       glCallList(light_list);
        glTranslatef(0,0,-WINDOW_ZOOM);
          /* place bitmap background */
        glPushMatrix();
            glRasterPos3f(-3.0,-3.0,0);
            glCallList(background);
        glPopMatrix();
              /* center screen on center of mass */
            glTranslatef(-CenterOfMass.x,-CenterOfMass.y,-CenterOfMass.z);

              /* Call display lists for each gravity object */
            for(i=0;i<MasterList.GetNumObjects();i++){
                current_obj = MasterList.GetObject(i);   
                glPushMatrix();
                      /* Translate to object position */
                    glTranslatef(current_obj->GetPosition().x,
                                 current_obj->GetPosition().y,
                                 current_obj->GetPosition().z);

                      /* Call object display list */
                    glCallList(current_obj->displayList);
                glPopMatrix();

                  /* run particle fountain for object */
                particleFountain(current_obj,deltaTime);
            }

    glPopMatrix();
    glFlush();
}

/************************************************************
*    mainReshape
*    Usage: mainReshape((int) width, (int) height);
*    Purpose: performs the window resizing
*    Returns: Nothing
*************************************************************/
void mainReshape(int w,int h){
    windim = (GLdouble) h/ (GLdouble) w;
    glViewport(0,0,(GLsizei) w,(GLsizei) h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-1.0,1.0,-windim,windim,1.0,41.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

/************************************************************
*    mainIdle
*    Usage: mainIdle(GLdouble);
*    Purpose: controls the operation of the gravity objects
*             calculates position and collisions
*    Returns: Nothing
*************************************************************/
void mainIdle(GLdouble deltaTime){
    GLuint i,j;
    Vector force;                          /* force vector */
    Vector acceleration;                   /* acceleration vector */
    Vector zeroVector;                     /* zero vector */
    Vertex tempPosition,tempPosition2;     /* interpolated collision positions */
    GLdouble seperation_1, seperation_2;     /* object seperations */
    GravityObject *current_obj,*next_obj;

      /* Zero Vector */
    zeroVector.x = 0;zeroVector.y = 0;zeroVector.z = 0;zeroVector.w = 0;

      /* In order to allow for simultaneous collision of 3 or */
      /* more objects, a temporary set which doesn't change */
      /* The calculation results are summed into the original set */
    for(i=0;i<MasterList.GetNumObjects();i++)
        TempObject[i] = *(MasterList.GetObject(i));

    for(i=0;i<MasterList.GetNumObjects();i++){
          /* zero acceleration vector */
        current_obj = MasterList.GetObject(i);
        current_obj->SetAcceleration(zeroVector);
        
        for(j=0;j<MasterList.GetNumObjects();j++){
            next_obj = MasterList.GetObject(j);
            if(i!=j){
                  /* Check if a collision between two objects has happened */
                  /* This allows us to avoid the calculations.  If one collsion */
                  /* happened, the opposite collistion must have happened */
                if(CollisionMatrix[i][j] == TRUE){
                      /* calculate the after-collision velocity */
                    current_obj->SetVelocity(current_obj->GetVelocity() +
                          calculateCollision(&TempObject[i],&TempObject[j]));
                }
                  /* check for a collision */
                else{
                      /* calculate if two objects are closer than the */
                      /* sum of there radius */
                    seperation_1 = (TempObject[i].GetPosition() -
                                    TempObject[j].GetPosition()).magnitude();

                    if(seperation_1 <= (TempObject[i].GetRadius() + 
                                        TempObject[j].GetRadius())){

                          /* interpolate object a small distance forware */
                        tempPosition = calculatePosition(TempObject[i].GetVelocity(),
                                                         TempObject[i].GetPosition(),
                                                         .0001);
                          /* interpolate other object */
                        tempPosition2 = calculatePosition(TempObject[j].GetVelocity(),
                                                          TempObject[j].GetPosition(),
                                                          .0001);
                          /* calculate distance between the new points */
                        seperation_2 = (tempPosition - tempPosition2).magnitude();
                          /* if the new distance is smaller than the first, objects */
                          /* are closing, and thus generate a collision */
                        if(seperation_2 < seperation_1 || CollisionMatrix[i][j] == TRUE){
                            CollisionMatrix[i][j] = TRUE;
                            CollisionMatrix[j][i] = TRUE;
                              /* calculate after collision velocity */
                            current_obj->SetVelocity(current_obj->GetVelocity() +
                                    calculateCollision(&TempObject[i],&TempObject[j]));
                        }
                    }
                }
                  /* calculate the gravitational force between all the objects */
                force = calculateGravityForce(current_obj,
                                              next_obj);
                  /* calculate the acceleration for each object */
                acceleration = calculateAcceleration(force,
                                                     current_obj->GetMass(),
                                                     deltaTime);

                  /* sum all the accelerations in order to get total gravity */
                  /* effect for each object */
                current_obj->SetAcceleration(current_obj->GetAcceleration() +
                                                         acceleration);
            }
        }
    }

      /* The position and velocity calculations must be done seperate */
      /* in order not to affect collision and gravity calculations */
    for(i=0;i<MasterList.GetNumObjects();i++){
          /* Calculate the velocity of each object */
        current_obj = MasterList.GetObject(i);
        current_obj->SetVelocity(calculateVelocity(current_obj->GetAcceleration(),
                                  current_obj->GetVelocity(),
                                  deltaTime));

          /* calculate position */
        current_obj->SetPosition(calculatePosition(current_obj->GetVelocity(),
                                  current_obj->GetPosition(),
                                  deltaTime));
    }
      /* clear collision matrix */
    for(i=0;i<MasterList.GetNumObjects();i++){
        for(j=0;j<MasterList.GetNumObjects();j++){
            CollisionMatrix[i][j] = FALSE;
        }
    }
}

/************************************************************
*    mainKey
*    Usage: mainKey((unsigned char) key,(int) X position, (int) Y position);
*    Purpose: Handles key strokes while mouse is in window
*    Returns: Nothing
*************************************************************/
void mainKey(unsigned char key,int,int){
    switch(key){
          // Escape key
        case 27:
            exit(0);
            break;
    }
}


/************************************************************
*    clear_hook
*    Usage: clear_hook();
*    Purpose: clears memory upon exit
*    Returns: Nothing
*************************************************************/
void clear_hook(void){
    GLuint i,j;
    GravityObject *current_obj;
    for(i=0;i<MasterList.GetNumObjects();i++){
        current_obj = MasterList.GetObject(i);
        for(j=0;j<current_obj->numParticles;j++)
            free(current_obj->particleList[j]);
        free(current_obj);
    }
    DestroyGLWindow();
}

/************************************************************
*    GravityList class
*    Purpose: Contains the global information for the set of 
*             gravity objects used.
*************************************************************/

/************************************************************
*    GravityList constructor
*    Usage: GravityList list;
*    Purpose: Creates the gravity list object with default of
*             5 gravity objects max.
*************************************************************/

GravityList::GravityList(void){
    NumObjects = 0;
    MaxObjects = 5;
}

/************************************************************
*    GravityList constructor
*    Usage: GravityList list(5);
*    Purpose: Creates the gravity list object with the spec. num
*             of gravity objects max.
*************************************************************/

GravityList::GravityList(GLuint num_objects){
    NumObjects = 0;
    if(num_objects < MAX_GRAVITY_OBJ)
        MaxObjects = num_objects;
    else
        MaxObjects = MAX_GRAVITY_OBJ;
}

/************************************************************
*    GravityList destructor
*    Purpose: Destroys the gravity list object
*************************************************************/

GravityList::~GravityList(void){
}

/************************************************************
*    GravityList accessor
*    Usage: list.GetNumObjects();
*    Purpose: returns the number of objects currently in the
*             object list;
*    Returns: Number of objects currently in the object list
*************************************************************/

GLuint GravityList::GetNumObjects(void){
    return NumObjects;
}

/************************************************************
*    GravityList accessor
*    Usage: list.GetMaxObjects();
*    Purpose: returns the max number of objects in the
*             object list;
*    Returns: the max objects held in the list
*************************************************************/

GLuint GravityList::GetMaxObjects(void){
    return MaxObjects;
}

/************************************************************
*    GravityList function
*    Usage: list.AddObjects(GraviyObject *object);
*    Purpose: Adds the object to the object list
*    Returns: success or error
*************************************************************/

int GravityList::AddObject(GravityObject *object){
    if(NumObjects >= MaxObjects)
        return -1;
    GravityObjectList[NumObjects] = object;
    NumObjects++;
    return 0;
}

/************************************************************
*    GravityList accessor
*    Usage: list.GetObject( int object_num);
*    Purpose: Returns the object in the object list
*    Returns: pointer to the Gravity object
*************************************************************/

GravityObject *GravityList::GetObject(GLuint object_num){
    if(object_num >= NumObjects)
        return NULL;
    return GravityObjectList[object_num];
}
