/************************************************************
*    win_funcs.c
*
*    Functions:
*        GetOpenGLPalette
*        SetDCPixelFormat
*        WinMain
*        WndProc
*************************************************************/
/************************************************************
*    Code created and written by Chris Hegarty
*    Feel free to use any portion of the code,
*    but I would appreciate credit if you use it.
*************************************************************/

#include <windows.h>
#include <gl/gl.h>
#include <stdlib.h>
#include "phys_eng.h"
#include "win_funcs.h"
#include "exit_handler.h"

static LPCTSTR lpszAppName = "Newton's Fountains";
static HINSTANCE hInstance;
static HPALETTE hRetPal;
static HPALETTE hPalette;
static HDC hDC;
static HGLRC hRC;
static HWND MainWindow;

HPALETTE GetOpenGLPalette(HDC hDC)
    {
    PIXELFORMATDESCRIPTOR pfd;
    LOGPALETTE *pPal;
    int nPixelFormat;
    int nColors;
    int i;
    BYTE RedRange,GreenRange,BlueRange;

    hRetPal = NULL;

    nPixelFormat = GetPixelFormat(hDC);
    DescribePixelFormat(hDC, nPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

    if(!(pfd.dwFlags & PFD_NEED_PALETTE))
        return NULL;

    nColors = 1 << pfd.cColorBits;
    pPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE) +nColors*sizeof(PALETTEENTRY));


    pPal->palVersion = 0x300;
    pPal->palNumEntries = nColors;

    RedRange = (1 << pfd.cRedBits) -1;
    GreenRange = (1 << pfd.cGreenBits) - 1;
    BlueRange = (1 << pfd.cBlueBits) -1;

    for(i = 0; i < nColors; i++){
        pPal->palPalEntry[i].peRed = (i >> pfd.cRedShift) & RedRange;
        pPal->palPalEntry[i].peRed = (unsigned char)(
            (double) pPal->palPalEntry[i].peRed * 255.0 / RedRange);
        pPal->palPalEntry[i].peGreen = (i >> pfd.cGreenShift) & GreenRange;
        pPal->palPalEntry[i].peGreen = (unsigned char)(
            (double)pPal->palPalEntry[i].peGreen * 255.0 / GreenRange);
        pPal->palPalEntry[i].peBlue = (i >> pfd.cBlueShift) & BlueRange;
        pPal->palPalEntry[i].peBlue = (unsigned char)(
            (double)pPal->palPalEntry[i].peBlue * 255.0 / BlueRange);
        pPal->palPalEntry[i].peFlags = (unsigned char) NULL;
    }

    hRetPal = CreatePalette(pPal);
    SelectPalette(hDC,hRetPal,FALSE);
    RealizePalette(hDC);

    free(pPal);
    return hRetPal;
}

void SetDCPixelFormat(HDC hDC){
    int nPixelFormat;

    static PIXELFORMATDESCRIPTOR pfd = {
        sizeof(PIXELFORMATDESCRIPTOR),
        1,
        PFD_DRAW_TO_WINDOW |                    // Draw to Window (not to bitmap)
        PFD_SUPPORT_OPENGL |                    // Support OpenGL calls in window
        PFD_DOUBLEBUFFER,                       // Double buffered
        PFD_TYPE_RGBA,                          // RGBA Color mode
        24,                                     // Want 24bit color
        0,0,0,0,0,0,                            // Not used to select mode
        0,0,                                    // Not used to select mode
        0,0,0,0,0,                              // Not used to select mode
        32,                                     // Size of depth buffer
        0,                                      // Not used to select mode
        0,                                      // Not used to select mode
        PFD_MAIN_PLANE,                         // Draw in main plane
        0,                                      // Not used to select mode
        0,0,0 };                                // Not used to select mode

    nPixelFormat = ChoosePixelFormat(hDC, &pfd);

    SetPixelFormat(hDC, nPixelFormat, &pfd);
}

int APIENTRY WinMain(HINSTANCE hInst,HINSTANCE,LPSTR,int){
    MSG             msg;
    WNDCLASS        wc;
    int x,y;

    setup_exit();

    hInstance = hInst;
    hPalette = NULL;

    wc.style                = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    wc.lpfnWndProc          = (WNDPROC) WndProc;
    wc.cbClsExtra           = 0;
    wc.cbWndExtra           = 0;
    wc.hInstance            = hInstance;
    wc.hIcon                = NULL;
    wc.hCursor              = NULL;
    wc.hbrBackground        = NULL;
    wc.lpszMenuName         = NULL;
    wc.lpszClassName        = lpszAppName;

    if(RegisterClass(&wc) == 0){
        MessageBox(NULL,"Could not register window","Error",MB_OK);
        return FALSE;
    }

    MainWindow= CreateWindow(
                lpszAppName,
                lpszAppName,
                WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                0, 0,
                800, 600,
                NULL,
                NULL,
                hInstance,
                NULL);

    if(MainWindow == NULL){
        MessageBox(NULL,"Could not create window","Error",MB_OK);
        return FALSE;
    }

    ShowWindow(MainWindow,SW_SHOW);
    
    DEVMODE test_dev;
	EnumDisplaySettings(NULL,0,&test_dev);
	x = GetSystemMetrics(SM_CXFULLSCREEN);
    y = GetSystemMetrics(SM_CYFULLSCREEN);
	if(x != 800 || y != 600){
		test_dev.dmPelsWidth = 800;
		test_dev.dmPelsHeight = 600;
		test_dev.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;
		if(ChangeDisplaySettings(&test_dev,0) != DISP_CHANGE_SUCCESSFUL){
		    MessageBox(NULL,"Could not change display settings","Error",MB_OK);
		    return FALSE;
		}
	}
	
    UpdateWindow(MainWindow);

	ShowCursor(FALSE);

    while( GetMessage(&msg, NULL, 0, 0)){
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    return msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hWnd,UINT message,
                        WPARAM wParam,LPARAM lParam){
    static RECT rect;
    
    switch (message)
        {

        case WM_CREATE:
            hDC = GetDC(hWnd);
            SetDCPixelFormat(hDC);
            hPalette = GetOpenGLPalette(hDC);
            hRC = wglCreateContext(hDC);
            wglMakeCurrent(hDC, hRC);
            windowInit();
            physEngInit();
            GetClientRect(hWnd,&rect);
            break;
        case WM_DESTROY:
            DestroyGLWindow();
            break;
        case WM_SIZE:
            mainReshape(LOWORD(lParam), HIWORD(lParam));
            break;
        case WM_PAINT:
            mainDisplay();
            SwapBuffers(hDC);
            ValidateRect(hWnd,&rect);
            InvalidateRect(hWnd,NULL,FALSE);
            break;
        case WM_KEYDOWN:
            mainKey(wParam,0,0);
            break;
        case WM_SYSCOMMAND:
			switch (wParam){
				case SC_SCREENSAVE:
					return 0;
			}
			return (DefWindowProc(hWnd, message, wParam, lParam));
        case WM_QUERYNEWPALETTE:
            if(hPalette){
                int nRet;
                SelectPalette(hDC, hPalette, FALSE);
                nRet = RealizePalette(hDC);
                InvalidateRect(hWnd,NULL,FALSE);
                return nRet;
            }
            break;
        case WM_PALETTECHANGED:
            if((hPalette != NULL) && ((HWND)wParam != hWnd)){
                SelectPalette(hDC,hPalette,FALSE);
                RealizePalette(hDC);
                UpdateColors(hDC);
                return 0;
            }
            break;
        default:
            return (DefWindowProc(hWnd, message, wParam, lParam));
    }
    return 0;
}

void DestroyGLWindow(void){
    wglMakeCurrent(hDC,NULL);
    wglDeleteContext(hRC);
    ReleaseDC(MainWindow,hDC);
    if(hPalette != NULL)
        DeleteObject(hPalette);
	ChangeDisplaySettings(NULL,0);
    PostQuitMessage(0);
}